@@ -58,38 +58,58 @@ class TestModels(unittest.TestCase):
         channel_first=False,
         **kwargs,
     ):
-        self.assertEqual(vision_tower.model_type, model_type)
-
-        batch_size = 1
-        if model_type == "qwen2_5_vl":
-            input_tensor = mx.random.uniform(shape=(image_size[0], image_size[1]))
-        else:
-            shape = (
-                (batch_size, num_channels, image_size[0], image_size[1])
-                if channel_first
-                else (batch_size, image_size[0], image_size[1], num_channels)
+        for t in [mx.float32, mx.float16]:
+            vision_tower.update(
+                tree_map(lambda p: p.astype(t), vision_tower.parameters())
             )
-            input_tensor = mx.random.uniform(shape=shape)
+            self.assertEqual(vision_tower.model_type, model_type)
 
-        if "image_masks" in inspect.signature(vision_tower.__call__).parameters:
-            input_tensor = input_tensor.transpose(0, 3, 1, 2)
-            image_masks = mx.ones((batch_size, num_channels, image_size[0]))
-            kwargs["image_masks"] = image_masks
+            if model_type == "llama4_vision_model":
+                vision_hidden_size = kwargs.pop(
+                    "projector_output_dim", vision_hidden_size
+                )
 
-        if (
-            "output_hidden_states"
-            in inspect.signature(vision_tower.__call__).parameters
-        ):
-            hidden_states = vision_tower(
-                input_tensor, output_hidden_states=True, **kwargs
-            )
-        else:
-            hidden_states = vision_tower(input_tensor, **kwargs)
+            batch_size = kwargs.pop("batch_size", 1)
+            if model_type == "qwen2_5_vl":
+                input_tensor = mx.random.uniform(shape=(image_size[0], image_size[1]))
+            else:
+                shape = (
+                    (batch_size, num_channels, image_size[0], image_size[1])
+                    if channel_first
+                    else (batch_size, image_size[0], image_size[1], num_channels)
+                )
+                input_tensor = mx.random.uniform(shape=shape)
 
-        # Check vision hidden feature layer's shape matches the expected hidden size
-        self.assertEqual(
-            hidden_states[vision_feature_layer].shape[-1], vision_hidden_size
-        )
+            if "image_masks" in inspect.signature(vision_tower.__call__).parameters:
+                input_tensor = input_tensor.transpose(0, 3, 1, 2)
+                image_masks = mx.ones((batch_size, num_channels, image_size[0]))
+                kwargs["image_masks"] = image_masks
+
+            input_tensor = input_tensor.astype(t)
+
+            if (
+                "output_hidden_states"
+                in inspect.signature(vision_tower.__call__).parameters
+            ):
+                hidden_states = vision_tower(
+                    input_tensor, output_hidden_states=True, **kwargs
+                )
+            else:
+                hidden_states = vision_tower(input_tensor, **kwargs)
+
+            hidden_states = hidden_states[vision_feature_layer]
+
+            # Check vision hidden feature layer's shape matches the expected hidden size
+            if channel_first:
+                if model_type == "llama4_vision_model":
+
+                    self.assertEqual(hidden_states.shape[1], vision_hidden_size)
+                else:
+                    self.assertEqual(hidden_states.shape[1], vision_hidden_size)
+            else:
+                self.assertEqual(hidden_states.shape[-1], vision_hidden_size)
+
+            self.assertEqual(hidden_states.dtype, t)
 
     def test_llava_bunny(self):
         from mlx_vlm.models import llava_bunny
@@ -407,6 +427,63 @@ class TestModels(unittest.TestCase):
             (config.vision_config.image_size, config.vision_config.image_size),
         )
 
+    def test_internvl_chat(self):
+        from mlx_vlm.models import internvl_chat
+
+        test_config = internvl_chat.TextConfig(
+            model_type="qwen2",
+            hidden_size=3584,
+            num_hidden_layers=5,
+            intermediate_size=18944,
+            num_attention_heads=28,
+            rms_norm_eps=1e-6,
+            max_window_layers=16,
+            use_sliding_window=False,
+            vocab_size=151674,
+            num_key_value_heads=4,
+            rope_theta=1000000.0,
+            rope_scaling={"factor": 2.0, "rope_type": "dynamic", "type": "dynamic"},
+            hidden_act="silu",
+            max_position_embeddings=32768,
+        )
+
+        vision_config = internvl_chat.VisionConfig(
+            model_type="intern_vit_6b",
+            num_hidden_layers=5,
+            hidden_size=1152,
+            intermediate_size=4304,
+            num_attention_heads=16,
+            image_size=384,
+            patch_size=14,
+            num_channels=3,
+            layer_norm_eps=1e-6,
+        )
+
+        config = internvl_chat.ModelConfig(
+            text_config=test_config,
+            vision_config=vision_config,
+            model_type="internvl_chat",
+            ignore_index=-100,
+            image_token_index=151667,
+        )
+
+        model = internvl_chat.Model(config)
+
+        self.language_test_runner(
+            model.language_model,
+            config.text_config.model_type,
+            config.text_config.vocab_size,
+            config.text_config.num_hidden_layers,
+        )
+
+        self.vision_test_runner(
+            model.vision_model,
+            config.vision_config.model_type,
+            config.vision_config.hidden_size,
+            config.vision_config.num_channels,
+            (config.vision_config.image_size, config.vision_config.image_size),
+        )
+
     def test_paligemma(self):
         from mlx_vlm.models import paligemma
 
@@ -1064,6 +1141,143 @@ class TestModels(unittest.TestCase):
             (config.vision_config.image_size, config.vision_config.image_size),
         )
 
+    def test_llama4(self):
+        from mlx_vlm.models import llama4
+
+        text_config = llama4.TextConfig(
+            model_type="llama4_text",
+            hidden_size=5120,
+            num_hidden_layers=3,
+            intermediate_size=8192,
+            intermediate_size_mlp=16384,
+            num_attention_heads=40,
+            num_key_value_heads=8,
+            rms_norm_eps=1e-05,
+            vocab_size=32000,
+            attention_chunk_size=8192,
+            attention_dropout=0.0,
+            head_dim=128,
+            hidden_act="silu",
+            attention_bias=False,
+        )
+        vision_config = llama4.VisionConfig(
+            model_type="llama4_vision_model",
+            image_size=336,
+            patch_size=14,
+            num_channels=3,
+            num_hidden_layers=3,
+            hidden_size=1408,
+            intermediate_size=5632,
+            num_attention_heads=16,
+            norm_eps=1e-05,
+            initializer_range=0.02,
+            pixel_shuffle_ratio=0.5,
+            projector_input_dim=4096,
+            projector_output_dim=4096,
+            projector_dropout=0.0,
+            vision_output_dim=4096,
+            rope_theta=10000,
+            vision_feature_layer=-1,
+            vision_feature_select_strategy="default",
+        )
+        config = llama4.ModelConfig(
+            text_config=text_config,
+            vision_config=vision_config,
+            model_type="llama4",
+        )
+        model = llama4.Model(config)
+
+        self.language_test_runner(
+            model.language_model,
+            config.text_config.model_type,
+            config.text_config.vocab_size,
+            config.text_config.num_hidden_layers,
+        )
+
+        self.vision_test_runner(
+            model.vision_model,
+            config.vision_config.model_type,
+            config.vision_config.hidden_size,
+            config.vision_config.num_channels,
+            (config.vision_config.image_size, config.vision_config.image_size),
+            channel_first=True,
+            projector_output_dim=config.vision_config.projector_output_dim,
+        )
+
+    def test_kimi_vl(self):
+        from mlx_vlm.models import kimi_vl
+
+        text_config = kimi_vl.TextConfig()
+        vision_config = kimi_vl.VisionConfig()
+        config = kimi_vl.ModelConfig(
+            text_config=text_config,
+            vision_config=vision_config,
+            model_type="kimi_vl",
+        )
+        model = kimi_vl.Model(config)
+
+        self.language_test_runner(
+            model.language_model,
+            config.text_config.model_type,
+            config.text_config.vocab_size,
+            config.text_config.num_hidden_layers,
+        )
+
+        self.vision_test_runner(
+            model.vision_tower,
+            config.vision_config.model_type,
+            config.vision_config.hidden_size,
+            config.vision_config.num_channels,
+            (config.vision_config.patch_size, config.vision_config.patch_size),
+            grid_thw=mx.array(
+                [[20, 28], [22, 28]], dtype=mx.int64
+            ),  # image temporals shape (num_images, 3)
+            batch_size=1176,
+            vision_feature_layer=-1,
+        )
+
+    def test_gemma3(self):
+        from mlx_vlm.models import gemma3
+
+        text_config = gemma3.TextConfig(
+            model_type="gemma3",
+            hidden_size=2048,
+            num_hidden_layers=18,
+            intermediate_size=16384,
+            num_attention_heads=8,
+            rms_norm_eps=1e-6,
+            vocab_size=257216,
+        )
+        vision_config = gemma3.VisionConfig(
+            model_type="gemma3",
+            image_size=224,
+            patch_size=14,
+            num_channels=3,
+            num_hidden_layers=18,
+            hidden_size=2048,
+            intermediate_size=16384,
+            num_attention_heads=8,
+        )
+        config = gemma3.ModelConfig(
+            text_config=text_config, vision_config=vision_config, model_type="gemma3"
+        )
+        model = gemma3.Model(config)
+
+        self.language_test_runner(
+            model.language_model,
+            config.text_config.model_type,
+            config.text_config.vocab_size,
+            config.text_config.num_hidden_layers,
+        )
+
+        self.vision_test_runner(
+            model.vision_tower,
+            config.vision_config.model_type,
+            config.vision_config.hidden_size,
+            config.vision_config.num_channels,
+            (config.vision_config.image_size, config.vision_config.image_size),
+        )
+
 
 if __name__ == "__main__":
     unittest.main()
