@@ -16,7 +16,7 @@ import numpy as np
 import requests
 from PIL import Image
 
-from .utils import generate, load, load_image
+from .utils import generate, load, load_image, process_inputs_with_fallback
 
 # This is a beta version of the video generation script.
 # It is not fully tested and may not work as expected.
@@ -513,7 +513,7 @@ def main():
             images=image_inputs,
             videos=video_inputs,
             padding=True,
-            return_tensors="np",
+            return_tensors="pt",
         )
 
         input_ids = mx.array(inputs["input_ids"])
@@ -569,13 +569,22 @@ def main():
             processor.image_processor.do_image_splitting = False
 
         # Process inputs
-        inputs = processor(
-            text=text, images=[img for img in frames], return_tensors="np"
+        inputs = process_inputs_with_fallback(
+            processor,
+            images=[img for img in frames],
+            prompts=text,
         )
 
         input_ids = mx.array(inputs["input_ids"])
         pixel_values = mx.array(inputs["pixel_values"])
         mask = mx.array(inputs["attention_mask"])
+        for key, value in inputs.items():
+            if key not in [
+                "input_ids",
+                "pixel_values",
+                "attention_mask",
+            ] and not isinstance(value, (str, list)):
+                kwargs[key] = mx.array(value)
 
     logger.info("\033[32mGenerating response...\033[0m")
 
